(*
high-level commands and declarations

we need a better name
*)

signature TERM_FUN =
sig

  val register_datatypes : string list -> theory -> theory

  val setup : theory -> theory
end

structure Term_Fun : TERM_FUN =
struct

fun register_datatypes tnames =
  Embed.define_embedding (hd tnames)
  #> fold Trs_Data.declare_constrs tnames
  #> fold Isafor_Tactics.add_constr_sims tnames

open Conv


fun define_trs prove_sim raw_simps thy =
  let
    val ctxt = Proof_Context.init_global thy
    val simps = Attrib.eval_thms ctxt raw_simps
    val conv : conv = (concl_conv ~1 o arg_conv o arg_conv)
      (Simplifier.rewrite (HOL_basic_ss addsimps (Trs_Data.Subst.get ctxt)))
    val simps' = simps |> map (fconv_rule conv)
                 
    fun add_sim_rule thy =
      let
        val sim_rule =
          Isafor_Tactics.derive_sim_lemma_any (Proof_Context.init_global thy) simps'
          |> Thm.close_derivation
      in
        Context.theory_map (Trs_Data.add_sim sim_rule) thy
      end
  in
    thy
    |> Trs_Data.define_trs simps'
    |> (if prove_sim then add_sim_rule else I)
  end

fun define_dps raw_simps thy =
  let
    val ctxt = Proof_Context.init_global thy
    val simps = Attrib.eval_thms ctxt raw_simps
    val conv : conv = (concl_conv ~1 o arg_conv o arg_conv)
      (Simplifier.rewrite (HOL_basic_ss addsimps (Trs_Data.Subst.get ctxt)))
    val simps' = simps |> map (fconv_rule conv)
                 
  in
    thy
    |> Trs_Data.define_dps simps'
  end


fun ceta_import_proof_tac ctxt path i = 
  let 
    val thy = Proof_Context.theory_of ctxt
    val prf = File.read path
       |> Ceta_Parser.read_cpf
       |> cterm_of thy
  in 
    rtac @{thm Ceta_Checker.ceta_checker[of _ tf_QL]} i
    THEN PRIMITIVE (Drule.instantiate' [] [SOME prf])  
  end

fun gen_ceta_conv mk_conv =
  let
    val thy = @{theory}
    val constr_names = map fst o snd o fst o Code.get_type thy;
    val constructors = maps (constr_names o Sign.intern_type thy)
      ["trs_proof", "dp_proof", "lab", "redtriple_impl", "root_redtriple_impl", "term", "list",
       "nat", "ctxt", "option", "projL", "sl_inter","arithFun","order_tag", "af_entry", "sum",
       "join_info", "arctic", "arctic_rat", "pos", "nibble"];
  in
    mk_conv  
      ([@{const_name Ceta_Checker.ceta_checker}, 
        @{const_name Rat.Fract}, 
        @{const_name String.char.Char},
        @{const_name Int.nat},
        @{const_name HOL.Trueprop}] @ 
       constructors)
  end



val ceta_eval_conv = gen_ceta_conv (Code_Runtime.static_holds_conv @{theory})
val ceta_eval_tac = CONVERSION ceta_eval_conv THEN' rtac @{thm HOL.TrueI} 

val ceta_checker_simp_tac = Code_Simp.static_tac @{theory} NONE [@{const_name Ceta_Checker.ceta_checker}] 

val setup = 
(*  #> Datatype.interpretation (K register_datatypes) *)(* not working yet *)
  Method.setup @{binding trs_reduce}
    (Args.term >> (fn f => fn ctxt => SIMPLE_METHOD' (Isafor_Tactics.reduce_tac ctxt f)))
    "reduces a termination goal to the termination of a term rewrite system"
  #> Trs_Data.setup
  #> Method.setup @{binding ceta_import_proof}
    (Scan.lift Parse.path >> (fn path => fn ctxt => SIMPLE_METHOD' (ceta_import_proof_tac ctxt path)))
    "reads a termination proof from a file" 
  #> Method.setup @{binding dps_reduce}
    (Args.term >> (fn f => fn ctxt => SIMPLE_METHOD' (Isafor_Tactics.reduce_dp_tac ctxt f)))
    "reduces a termination goal to finiteness of a DP problem"
  #> Method.setup @{binding ceta_checker_simp}
    (Scan.succeed (fn _ => SIMPLE_METHOD' (ceta_checker_simp_tac)))
    "evaluates ceta_checker using the simplifier" 
  #> Method.setup @{binding ceta_eval}
    (Scan.succeed (fn _ => SIMPLE_METHOD' (ceta_eval_tac)))
    "evaluates ceta_checker using the evaluator" 

fun export_termination_problem path proof_state =
  let
    val {goal=goal_state,...} = Proof.goal proof_state
    val is_con = Proof.theory_of proof_state
               |> Trs_Data.is_constr
    val subgoal = hd (prems_of goal_state)
    val (qL,trsL) = case subgoal of 
      Const (@{const_name HOL.Trueprop}, _)
        $ (Const (@{const_name Abstract_Rewriting.SN_on}, _)
        $ (Const (@{const_name Q_Restricted_Rewriting.qrstep}, _)
        $ @{term True}
          $ (Const (@{const_name List.set}, _) $ qL)
          $ (Const (@{const_name List.set}, _) $ trsL))
        $ _ (*UNIV*)) => (qL,trsL)
    | _ => raise TERM ("Bad goal", [subgoal])
    val problem = Xml.mk_problem is_con qL trsL
  in
    File.write path (Xml.to_string problem)
  end


val parser =
  Scan.optional (Parse.$$$ "(" |-- (Parse.reserved "no_sim" >> K false) --| Parse.$$$ ")") true
  -- Parse_Spec.xthms1

val _ =
  Outer_Syntax.command @{command_spec "define_trs"} "defines term rewrite system corresponding to given simp rules"
    (parser >> (fn (sims,facts) => Toplevel.theory (define_trs sims facts)))

val _ =
  Outer_Syntax.command @{command_spec "define_dps"} "defines DP problem corresponding to given simp rules"
    (parser >> (fn (_,facts) => Toplevel.theory (define_dps facts)))


val _ =
  Outer_Syntax.command @{command_spec "export_termination_problem"} "write termination problem to file"
    (Parse.path >> (fn path => Toplevel.unknown_proof o
      Toplevel.keep (export_termination_problem path o Toplevel.proof_of)))


end
